/*
  Macro           MatchColorSet
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   Windows TSE v4.4 upwards
  Version         v1   26 Jul 2023

  This tool matches a set of 16 input colors to TSE's standard 16 colors.

  For example, you have found a set of 16 colors you like on the internet, and
  want to use them for TSE. TSE's set of 16 standard colors has color names
  which will probably not all match with the 16 colors you want to use.
  You can of course try to eye-ball it, but not everybody can, and where is the
  fun in a subjective match when an automated solution might be possible?

  This tool will calculate the difference of all 16 input colors with all
  16 standard colors, and will come up with a close match of the two sets.


  INSTALLATION

  Put this file in TSE's "mac" directory and compile it there, for example by
  opening it there in TSE and applying the Macro Compile menu.


  INPUT

  In a buffer in TSE, make its entire content 16 or 32 color values.
  (There is test data available further on in this documentation.)
  Format the buffer to one color per line.
  If you enter 16 colors, which is the norm, then the tool will match them
  with TSE's set of 16 standard colors.
  If you enter 32 colors, then the tool will match the first 16 "input" colors
  with the last 16 "reference" colors.
  Empty lines are allowed and ignored.
  Acceptable color value formats are, by example:
    0099ff
    #0099ff
    0x0099ff
    rgb(0 153 255)
    rgb(0, 153, 255)
  The tool ignores case and irrelevant spaces.


  USAGE

  While the input buffer is current, execute this tool.


  OUTPUT

  The output will be an HTML file, as I judged that to be the most portable
  format.
  You will be asked whether to open the HTML file with your default web
  browser.
  Then, for example, you can copy a web browser page into an email.


  KNOWN ERRORS

  1.  (Small)
  Even an average eye ball does not see all colors equally.
  For example, green is usually perceived as a brighter than average color,
  while "blue" is usually perceived as a darker than average color.
  This tool's algorithm simplifies the real world by treating the red, green
  and blue components of an RGB color as equally significant when calculating
  the difference between colors.
  So far I found, that in practice this tool's view being distorted by this is
  negligible.

  2.  (Significant)
  The tool is too short-sighted and aggressive in linking input colors to
  standard or reference colors.
  The tool works by first linking the two colors with the smallest difference.
  For the remaining colors it then again links the two colors with the smallest
  difference.
  And so on.
  The problem is that this can end with having to link badly mismatching
  colors for a color set for which that is avoidable.
  For example, not always linking the least differing colors can lead to a
  better overall matching color set.
  Literally: One test says, that a small as possible sum of all of a color
  set's linked differences is a good indicator of how well the color set
  matches as a whole.
  That said, an algorithm cannot examine 16! (16 factorial = 21 trillion)
  color sets.
  I only have a glimmer of an idea on how to maybe solve this.
  Maybe I will not pursue it, as this tool has low priority.


  TEST DATA

  Sammy's 1st request was to classify these colors:

    0x000000
    0x220077
    0x007722
    0x227788
    0x660033
    0x770088
    0x996600
    0x444455
    0x777799
    0x3322FF
    0x00FF33
    0x22EEFF
    0xFF0033
    0xFF33DD
    0xFFEE33
    0xFFFFFF

  Sammy's 2nd request was to classify these colors:

    RGB(255, 194,  10)
    RGB( 12, 123, 220)
    RGB(153,  79,   0)
    RGB(  0, 108, 209)
    RGB(225, 190, 106)
    RGB( 64, 176, 166)
    RGB(230,  97,   0)
    RGB( 93,  58, 155)
    RGB( 26, 255,  26)
    RGB( 75,   0, 146)
    RGB(254, 254,  98)
    RGB(211,  95, 183)
    RGB(  0,  90, 181)
    RGB(220,  50,  32)
    RGB( 26, 133, 255)
    RGB(212,  17,  89)


*/



/*
  T E C H N I CA L   I N F O


  CALCULATING THE DIFFERENCE BETWEEN TWO COLORS

  The algorithm pretends an RGB Color value is a tree-dimensional coordinate,
  and for two colors it calculates the distance between their coordinates.

  Even for average eyes this is not an exact representation of how eyes work,
  but my assumption is that this is a practical approximation.



  SQUARE ROOT

  TSE does not have a square root function, so the macro needs its own one.

  While the tool's speed is a factor, it is not a critical one, so I opted for
  a binary search, which in theory and in tests works ok.
*/



// You need to install my "Log" macro for debugging to work.
#define DEBUG FALSE



#ifdef LINUX
  Sorry, I have not made this tool LINUX compatible.
#else
  string  SLASH [1] = '\'
#endif

#if EDITOR_VERSION < 4400h
  Sorry, I have not made this tool compatible with TSE versions before 4.4 .
#endif



// Global constants and semi-constants

integer DATA_ID                   = 0
integer MACRO_ID                  = 0
string  MACRO_NAME [MAXSTRINGLEN] = ''
#define ROOT_MAXINT                 46340



// Global variables
string reference_type [9] = 'Reference'



// This datadef Must be exactly 16 lines with 2 spaces-separated columns
// like in the documentation of SetColorTableValue().

Datadef StandardTseColors
  '_BLACK_           0x000000'
  '_BLUE_            0x000080'
  '_GREEN_           0x008000'
  '_CYAN_            0x008080'
  '_RED_             0x800000'
  '_MAGENTA_         0x800080'
  '_BROWN_           0x808000'
  '_LIGHT_GRAY_      0xc0c0c0'
  '_DARK_GRAY_       0x808080'
  '_LIGHT_BLUE_      0x0000ff'
  '_LIGHT_GREEN_     0x00ff00'
  '_LIGHT_CYAN_      0x00ffff'
  '_LIGHT_RED_       0xff0000'
  '_LIGHT_MAG_       0xff00ff'
  '_YELLOW_          0xffff00'
  '_WHITE_           0xffffff'
end StandardTseColors



// Global variables
integer ok = TRUE


string proc get_tse_color_name(integer ref_nr)
  string result [MAXSTRINGLEN] = ''
  PushLocation()
  GotoBufferId(DATA_ID)
  GotoLine(ref_nr)
  result = Trim(GetToken(GetText(1, MAXSTRINGLEN), ' ', 1))
  PopLocation()
  return(result)
end get_tse_color_name


string proc get_color_string_value(string color_type, integer ref_nr)
  string  result [MAXSTRINGLEN] = ''
  integer result_b              = 0
  integer result_g              = 0
  integer result_r              = 0

  result_r = GetBufferInt(Format(color_type, '_', ref_nr, '_', 1), MACRO_ID)
  result_g = GetBufferInt(Format(color_type, '_', ref_nr, '_', 2), MACRO_ID)
  result_b = GetBufferInt(Format(color_type, '_', ref_nr, '_', 3), MACRO_ID)

  result = Format(result_r:2:'0':16,
                  result_g:2:'0':16,
                  result_b:2:'0':16,
                  ' - RGB(',
                  result_r:3, ', ' ,
                  result_g:3, ', ' ,
                  result_b:3, ')'  )
  result = Upper(result)

  return(result)
end get_color_string_value


string proc get_txt_color(string ref_color)
  integer average_component_value = 0
  string  result              [5] = ''

  average_component_value = ( Val(ref_color[1: 2], 16) +
                              Val(ref_color[3: 2], 16) +
                              Val(ref_color[5: 2], 16) ) / 3
  if average_component_value >= 128
    result = 'black'
  else
    result = 'white'
  endif

  return(result)
end get_txt_color


integer proc generate_html_output()
  string  html_fqn  [MAXSTRINGLEN] = ''
  string  new_color [MAXSTRINGLEN] = ''
  integer new_nr                   = 0
  string  ref_color [MAXSTRINGLEN] = ''
  integer ref_nr                   = 0
  string  txt_color            [5] = ''

  html_fqn = Format(RemoveTrailingSlash(GetEnvStr('TMP')), SLASH, 'TSE_',
                    MACRO_NAME, '.html')
  if not EditThisFile(html_fqn, _DONT_PROMPT_)
    Warn(MACRO_NAME; 'error: Could not open'; QuotePath(html_fqn), '.')
    ok = FALSE
    return(ok)
  endif

  EmptyBuffer()
  AddLine('<!doctype html>')
  AddLine('<html lang="en">')
  AddLine('  <head>')
  AddLine('    <meta charset="utf8">')
  AddLine('    <meta name="viewport" content="width=device-width, initial-scale=1.0">')
  AddLine('    <title>Match input colors to TSE colors or reference colors</title>')
  AddLine('  </head>')
  AddLine('  <body>')
  AddLine('    <table>')
  AddLine('      <tr>')
  if reference_type == 'TSE'
    AddLine('        <th>TSE Name</th>')
  endif
  AddLine('        <th>' + reference_type + ' Color</th>')
  AddLine('        <th>Closest Matched Input Color</th>')
  AddLine('      <tr>')

  for ref_nr = 1 to 16
    AddLine('      <tr>')
    if reference_type == 'TSE'
      AddLine('        <td>' + get_tse_color_name(ref_nr) + '</td>')
    endif
    ref_color = get_color_string_value('RefColor', ref_nr)
    txt_color = get_txt_color(ref_color)
    AddLine(Format('        <td style="color:',
                   txt_color,
                   ';background-color:#',
                   ref_color[1: 6],
                   '">',
                   ref_color,
                   '</td>'))
    new_nr    = GetBufferInt(Format('Match_', ref_nr), MACRO_ID)
    new_color = get_color_string_value('NewColor', new_nr)
    txt_color = get_txt_color(new_color)
    AddLine(Format('        <td style="color:',
                   txt_color,
                   ';background-color:#',
                   new_color[1: 6],
                   '">',
                   new_color,
                   '</td>'))
    AddLine('      </tr>')
  endfor

  AddLine('    </table>')
  AddLine('  </body>')
  AddLine('</html>')
  BegFile()
  UpdateDisplay()

  if YesNo('Open this HTML file in the default web browser?') == 1
    if SaveAs(html_fqn, _DONT_PROMPT_|_OVERWRITE_)
      StartPgm(html_fqn)
    else
      Warn(MACRO_NAME; 'error: Could not save'; QuotePath(html_fqn), '.')
      ok = FALSE
    endif
  endif

  return(ok)
end generate_html_output


integer proc find_first_set_match()
  integer distance              = 0
  integer lowest_distance       = 0
  integer lowest_new_nr         = 0
  integer lowest_ref_nr         = 0
  integer matched_color_counter = 0
  string  new_available     [3] = Chr(255) + Chr(255) + Chr(255) // All bits 1.
  integer new_nr                = 0
  string  ref_available     [3] = Chr(255) + Chr(255) + Chr(255) // All bits 1.
  integer ref_nr                = 0

  for matched_color_counter = 1 to 16
    lowest_distance = MAXINT
    for ref_nr = 1 to 16
      if GetBit(ref_available, ref_nr)
        for new_nr = 1 to 16
          if GetBit(new_available, new_nr)
            distance = GetBufferInt(Format('Distance_', ref_nr, '_', new_nr),
                                    MACRO_ID)
            if distance < lowest_distance
              lowest_distance = distance
              lowest_ref_nr   = ref_nr
              lowest_new_nr   = new_nr
            endif
          endif
        endfor
      endif
    endfor
    #if DEBUG
      ExecMacro(Format('Log write Matched';
                       lowest_ref_nr:2; 'to'; lowest_new_nr:2;
                       'with distance'; lowest_distance:4, '.'))
    #endif
    SetBufferInt(Format('Match_', lowest_ref_nr), lowest_new_nr, MACRO_ID)
    ClearBit(ref_available, lowest_ref_nr)
    ClearBit(new_available, lowest_new_nr)
  endfor

  return(ok)
end find_first_set_match


integer proc root(integer n)
  integer result        = 0
  integer result_middle = 0
  integer result_from   = 0
  integer result_to     = ROOT_MAXINT + 1
  if n > 0
    while result_from <> result_to - 1
      result_middle = (result_from + result_to) / 2
      if     result_middle * result_middle <= n
        result_from = result_middle
      else
        result_to   = result_middle
      endif
    endwhile
    result = result_from
  elseif n < 0
    Warn(MACRO_NAME; 'program error: Cannot root negative number.')
    ok =FALSE
  endif
  return(result)
end root


integer proc square(integer n)
  integer result = MAXINT

  if Abs(n) <= ROOT_MAXINT
    result = n * n
  else
    Warn(MACRO_NAME; 'program error: Cannot square too large integer.')
    ok = FALSE
  endif

  return(result)
end square


integer proc distance_between_points(integer point1_x,
                                     integer point1_y,
                                     integer point1_z,
                                     integer point2_x,
                                     integer point2_y,
                                     integer point2_z)
  integer result = 0

  result = root(square(point1_x - point2_x) +
                square(point1_y - point2_y) +
                square(point1_z - point2_z)   )

  return(result)
end distance_between_points

integer proc compute_all_distances()
  integer distance     = 0
  integer new_b        = 0
  integer new_color_nr = 0
  integer new_g        = 0
  integer new_r        = 0
  integer ref_b        = 0
  integer ref_color_nr = 0
  integer ref_g        = 0
  integer ref_r        = 0

  for ref_color_nr = 1 to 16
    ref_r = GetBufferInt(Format('RefColor_', ref_color_nr, '_1'), MACRO_ID)
    ref_g = GetBufferInt(Format('RefColor_', ref_color_nr, '_2'), MACRO_ID)
    ref_b = GetBufferInt(Format('RefColor_', ref_color_nr, '_3'), MACRO_ID)
    for new_color_nr = 1 to 16
      new_r = GetBufferInt(Format('NewColor_', new_color_nr, '_1'), MACRO_ID)
      new_g = GetBufferInt(Format('NewColor_', new_color_nr, '_2'), MACRO_ID)
      new_b = GetBufferInt(Format('NewColor_', new_color_nr, '_3'), MACRO_ID)
      distance = distance_between_points(ref_r, ref_g, ref_b,
                                         new_r, new_g, new_b)
      if ok
        #if DEBUG
          ExecMacro(Format('Log write Distance from ref'; ref_color_nr:2; '(',
                           ref_r:3; ref_g:3; ref_b:3,
                           ') to new'; new_color_nr:2; '(',
                           new_r:3; new_g:3; new_b:3,
                           ') is'; distance:4, '.'))
        #endif
        SetBufferInt(Format('Distance_', ref_color_nr, '_', new_color_nr),
                     distance, MACRO_ID)
      else
        break
      endif
    endfor
    if not ok
      break
    endif
  endfor

  return(ok)
end compute_all_distances


integer proc read_color_input()
  integer color_counter                  = 0
  integer color_nr                       = 0
  string  color_string    [MAXSTRINGLEN] = ''
  string  color_type                 [8] = 'NewColor'
  integer component_nr                   = 0
  string  component_value [MAXSTRINGLEN] = ''
  BegFile()
  repeat
    color_string = Lower(Trim(GetText(1, CurrLineLen())))
    if color_string <> ''
      color_counter = color_counter + 1
      color_nr      = color_nr      + 1
      if color_counter == 17
        color_nr   = 1
        color_type = 'RefColor'
      endif
      if color_string[1: 3] == 'rgb'
        color_string = StrReplace('rgb', color_string, '' )
        color_string = StrReplace('('  , color_string, '' )
        color_string = StrReplace(')'  , color_string, '' )
        color_string = StrReplace(','  , color_string, ' ')
        for component_nr = 1 to 3
          component_value = GetToken(color_string, ' ', component_nr)
          if isDigit(component_value)
            SetBufferInt(Format(color_type, '_', color_nr, '_', component_nr),
                         Val(component_value),
                         MACRO_ID)
          else
            Warn(MACRO_NAME; 'input error: "', color_string, '".')
            ok = FALSE
            break
          endif
        endfor
      else
        color_string = StrReplace('0x', color_string, '')
        color_string = StrReplace('#' , color_string, '')
        for component_nr = 1 to 3
          component_value = color_string[component_nr * 2 - 1: 2]
          if isHexDigit(component_value)
            SetBufferInt(Format(color_type, '_', color_nr, '_', component_nr),
                         Val(component_value, 16),
                         MACRO_ID)
          else
            Warn(MACRO_NAME; 'input error: "', color_string, '".')
            ok = FALSE
            break
          endif
        endfor
      endif
    endif
  until not ok
     or color_counter > 32
     or not Down()

  if ok
    case color_counter
      when 16
        reference_type = 'TSE'
      when 32
        reference_type = 'Reference'
      otherwise
        Warn(MACRO_NAME; 'input error: Number of input colors is not 16 or 32.')
        ok = FALSE
    endcase
  endif

  if  ok
  and color_counter == 16
    PushLocation()
    GotoBufferId(DATA_ID)
    for color_nr = 1 to 16
      GotoLine(color_nr)
      color_string = GetToken(GetText(1, MAXSTRINGLEN), ' ', 2)
      for component_nr = 1 to 3
        component_value = color_string[component_nr * 2 - 1: 2]
        SetBufferInt(Format('RefColor_', color_nr, '_', component_nr),
                     Val(component_value, 16),
                     MACRO_ID)
      endfor
    endfor
    PopLocation()
  endif
  return(ok)
end read_color_input


proc WhenPurged()
  AbandonFile(DATA_ID)
  AbandonFile(MACRO_ID)
end WhenPurged


proc WhenLoaded()
  MACRO_NAME = SplitPath(CurrMacroFilename(), _NAME_)

  PushLocation()

  MACRO_ID   = CreateTempBuffer()
  ChangeCurrFilename(MACRO_NAME + ':BufferVariables',
                     _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)

  DATA_ID = CreateTempBuffer()
  ChangeCurrFilename(MACRO_NAME + ':StandardTseColors',
                     _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
  PushBlock()
  InsertData(StandardTseColors)
  PopBlock()
  lReplace('0x', '', 'gin')

  PopLocation()
end WhenLoaded


proc Main()
  #if DEBUG
    string log_fqn [MAXSTRINGLEN] = ''
    log_fqn = GetEnvStr('TMP') + SLASH + 'TSE_' + MACRO_NAME + '.log'
    EraseDiskFile(log_fqn)
    ExecMacro('Log open ' + QuotePath(log_fqn))
  #endif

  if read_color_input()
    if compute_all_distances()
      if find_first_set_match()
        generate_html_output()
      endif
    endif
  endif

  #if DEBUG
    ExecMacro('Log close')
  #endif

  PurgeMacro(MACRO_NAME)
end Main

